#pragma once

// Helper
#include "math/helper/helper.h"

// Imaginary
#include "math/imaginary/complex.h"
#include "math/imaginary/quaternion.h"

// Vector
#include "math/vector/vector2.h"
#include "math/vector/vector3.h"
#include "math/vector/vector4.h"

// Matrix
#include "math/matrix/matrix2x2.h"
#include "math/matrix/matrix3x3.h"
#include "math/matrix/matrix4x4.h"

// Triangle
#include "math/triangle/vertex.h"
#include "math/triangle/triangle.h"

// Ray
#include "math/ray/plane.h"
#include "math/ray/sphere.h"
#include "math/ray/ray.h"


namespace kl {
    // Complex
    template<typename T>
    constexpr Complex<T> abs(const Complex<T>& num)
    {
        return apply<2>(helper::abs<T>, num);
    }

    template<typename T>
    constexpr Complex<T> min(const Complex<T>& num1, const Complex<T>& num2)
    {
        return apply<2>(helper::min<T>, num1, num2);
    }

    template<typename T>
    constexpr Complex<T> max(const Complex<T>& num1, const Complex<T>& num2)
    {
        return apply<2>(helper::max<T>, num1, num2);
    }

    template<typename T>
    constexpr Complex<T> normalize(const Complex<T>& num)
    {
        return num * (T(1) / num.length());
    }

    template<typename T>
    constexpr Complex<T> inverse(const Complex<T>& num)
    {
        const T sqr_sum = num.r * num.r + num.i * num.i;
        if (sqr_sum == T(0)) {
            return {};
        }
        return { num.r / sqr_sum, -num.i / sqr_sum };
    }

    // Quaternion
    template<typename T>
    constexpr Quaternion<T> abs(const Quaternion<T>& num)
    {
        return apply<4>(helper::abs<T>, num);
    }

    template<typename T>
    constexpr Quaternion<T> min(const Quaternion<T>& num1, const Quaternion<T>& num2)
    {
        return apply<4>(helper::min<T>, num1, num2);
    }

    template<typename T>
    constexpr Quaternion<T> max(const Quaternion<T>& num1, const Quaternion<T>& num2)
    {
        return apply<4>(helper::max<T>, num1, num2);
    }

    template<typename T>
    constexpr Quaternion<T> normalize(const Quaternion<T>& num)
    {
        return num * (T(1) / num.length());
    }

    template<typename T>
    constexpr Quaternion<T> inverse(const Quaternion<T>& num)
    {
        return { num.w, -num.x, -num.y, -num.z };
    }

    // Vector2
    template<typename T>
    constexpr Vector2<T> abs(const Vector2<T>& vec)
    {
        return apply<2>(helper::abs<T>, vec);
    }

    template<typename T>
    constexpr Vector2<T> min(const Vector2<T>& vec1, const Vector2<T>& vec2)
    {
        return apply<2>(helper::min<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector2<T> max(const Vector2<T>& vec1, const Vector2<T>& vec2)
    {
        return apply<2>(helper::max<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector2<T> normalize(const Vector2<T>& vec)
    {
        return vec * (T(1) / vec.length());
    }

    template<typename T>
    constexpr T dot(const Vector2<T>& first, const Vector2<T>& second)
    {
        return first.x * second.x + first.y * second.y;
    }

    template<typename T>
    constexpr T angle(Vector2<T> first, Vector2<T> second, bool full)
    {
        first = normalize(first);
        second = normalize(second);

        T result = {};
        if (full) {
            result = std::atan2(first.x * second.y - first.y * second.x, first.x * second.x + first.y * second.y);
            result += (T) PI;
        }
        else {
            result = std::acos(dot(first, second));
        }
        return static_cast<T>(result * TO_DEGREES);
    }

    template<typename T>
    constexpr Vector2<T> rotate(const Vector2<T>& vec, T angle)
    {
        const T sin_a = (T) std::sin(angle * TO_RADIANS);
        const T cos_a = (T) std::cos(angle * TO_RADIANS);
        return { cos_a * vec.x - sin_a * vec.y, sin_a * vec.x + cos_a * vec.y };
    }

    template<typename T>
    constexpr Vector2<T> reflect(const Vector2<T>& vec, const Vector2<T>& point)
    {
        const Vector2<T> normalized = normalize(point);
        return vec - (normalized * (vec * normalized) * T(2));
    }

    // Vector3
    template<typename T>
    constexpr Vector3<T> abs(const Vector3<T>& vec)
    {
        return apply<3>(helper::abs<T>, vec);
    }

    template<typename T>
    constexpr Vector3<T> min(const Vector3<T>& vec1, const Vector3<T>& vec2)
    {
        return apply<3>(helper::min<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector3<T> max(const Vector3<T>& vec1, const Vector3<T>& vec2)
    {
        return apply<3>(helper::max<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector3<T> normalize(const Vector3<T>& vec)
    {
        return vec * (T(1) / vec.length());
    }

    template<typename T>
    constexpr T dot(const Vector3<T>& first, const Vector3<T>& second)
    {
        return first.x * second.x + first.y * second.y + first.z * second.z;
    }

    template<typename T>
    constexpr T angle(const Vector3<T>& first, const Vector3<T>& second)
    {
        const T result = std::acos(dot(normalize(first), normalize(second)));
        return static_cast<T>(result * TO_DEGREES);
    }

    template<typename T>
    constexpr Vector3<T> rotate(const Vector3<T>& vec, const Vector3<T>& axis, float angle)
    {
        const Quaternion<T> quat{ axis, angle };
        const Quaternion<T> inv_quat = inverse(quat);
        return quat * vec * inv_quat;
    }

    template<typename T>
    constexpr Vector3<T> reflect(const Vector3<T>& first, Vector3<T> normal)
    {
        normal = normalize(normal);
        return first - (normal * (first * normal) * T(2));
    }

    template<typename T>
    constexpr Vector3<T> cross(const Vector3<T>& first, const Vector3<T>& second)
    {
        return {
            first.y * second.z - first.z * second.y,
            first.z * second.x - first.x * second.z,
            first.x * second.y - first.y * second.x,
        };
    }

    // Vector4
    template<typename T>
    constexpr Vector4<T> abs(const Vector4<T>& vec)
    {
        return apply<4>(helper::abs<T>, vec);
    }

    template<typename T>
    constexpr Vector4<T> min(const Vector4<T>& vec1, const Vector4<T>& vec2)
    {
        return apply<4>(helper::min<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector4<T> max(const Vector4<T>& vec1, const Vector4<T>& vec2)
    {
        return apply<4>(helper::max<T>, vec1, vec2);
    }

    template<typename T>
    constexpr Vector4<T> normalize(const Vector4<T>& vec)
    {
        return vec * (T(1) / vec.length());
    }

    template<typename T>
    constexpr T dot(const Vector4<T>& first, const Vector4<T>& second)
    {
        return first.x * second.x + first.y * second.y + first.z * second.z + first.w * second.w;
    }

    template<typename T>
    constexpr T angle(const Vector4<T>& first, const Vector4<T>& second)
    {
        const T result = std::acos(dot(normalize(first), normalize(second)));
        return static_cast<T>(result * TO_DEGREES);
    }

    // Matrix2x2
    template<typename T>
    constexpr Matrix2x2<T> abs(const Matrix2x2<T>& mat)
    {
        return apply<4>(helper::abs<T>, mat);
    }

    template<typename T>
    constexpr Matrix2x2<T> min(const Matrix2x2<T>& mat1, const Matrix2x2<T>& mat2)
    {
        return apply<4>(helper::min<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix2x2<T> max(const Matrix2x2<T>& mat1, const Matrix2x2<T>& mat2)
    {
        return apply<4>(helper::max<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix2x2<T> inverse(const Matrix2x2<T>& mat)
    {
        const T determinant = mat.determinant();

        Matrix2x2<T> result;
        result[0] = +mat[3];
        result[1] = -mat[1];
        result[2] = -mat[2];
        result[3] = +mat[0];
        return result * determinant;
    }

    template<typename T>
    constexpr Matrix2x2<T> transpose(const Matrix2x2<T>& mat)
    {
        Matrix2x2<T> result;
        for (int y = 0; y < 2; y++) {
            for (int x = 0; x < 2; x++) {
                result(x, y) = mat(y, x);
            }
        }
        return result;
    }

    // Matrix3x3
    template<typename T>
    constexpr Matrix3x3<T> abs(const Matrix3x3<T>& mat)
    {
        return apply<9>(helper::abs<T>, mat);
    }

    template<typename T>
    constexpr Matrix3x3<T> min(const Matrix3x3<T>& mat1, const Matrix3x3<T>& mat2)
    {
        return apply<9>(helper::min<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix3x3<T> max(const Matrix3x3<T>& mat1, const Matrix3x3<T>& mat2)
    {
        return apply<9>(helper::max<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix3x3<T> inverse(const Matrix3x3<T>& mat)
    {
        const T determinant = mat.determinant();

        Matrix3x3<T> result;
        result(0, 0) = (mat(1, 1) * mat(2, 2) - mat(2, 1) * mat(1, 2)) * determinant;
        result(0, 1) = (mat(0, 2) * mat(2, 1) - mat(0, 1) * mat(2, 2)) * determinant;
        result(0, 2) = (mat(0, 1) * mat(1, 2) - mat(0, 2) * mat(1, 1)) * determinant;
        result(1, 0) = (mat(1, 2) * mat(2, 0) - mat(1, 0) * mat(2, 2)) * determinant;
        result(1, 1) = (mat(0, 0) * mat(2, 2) - mat(0, 2) * mat(2, 0)) * determinant;
        result(1, 2) = (mat(1, 0) * mat(0, 2) - mat(0, 0) * mat(1, 2)) * determinant;
        result(2, 0) = (mat(1, 0) * mat(2, 1) - mat(2, 0) * mat(1, 1)) * determinant;
        result(2, 1) = (mat(2, 0) * mat(0, 1) - mat(0, 0) * mat(2, 1)) * determinant;
        result(2, 2) = (mat(0, 0) * mat(1, 1) - mat(1, 0) * mat(0, 1)) * determinant;
        return result;
    }

    template<typename T>
    constexpr Matrix3x3<T> transpose(const Matrix3x3<T>& mat)
    {
        Matrix3x3<T> result;
        for (int y = 0; y < 3; y++) {
            for (int x = 0; x < 3; x++) {
                result(x, y) = mat(y, x);
            }
        }
        return result;
    }

    // Matrix4x4
    template<typename T>
    constexpr Matrix4x4<T> abs(const Matrix4x4<T>& mat)
    {
        return apply<16>(helper::abs<T>, mat);
    }

    template<typename T>
    constexpr Matrix4x4<T> min(const Matrix4x4<T>& mat1, const Matrix4x4<T>& mat2)
    {
        return apply<16>(helper::min<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix4x4<T> max(const Matrix4x4<T>& mat1, const Matrix4x4<T>& mat2)
    {
        return apply<16>(helper::max<T>, mat1, mat2);
    }

    template<typename T>
    constexpr Matrix4x4<T> inverse(const Matrix4x4<T>& mat)
    {
        const T determinant = mat.determinant();

        const T a_2323 = mat(2, 2) * mat(3, 3) - mat(3, 2) * mat(2, 3);
        const T a_1323 = mat(1, 2) * mat(3, 3) - mat(3, 2) * mat(1, 3);
        const T a_1223 = mat(1, 2) * mat(2, 3) - mat(2, 2) * mat(1, 3);
        const T a_0323 = mat(0, 2) * mat(3, 3) - mat(3, 2) * mat(0, 3);
        const T a_0223 = mat(0, 2) * mat(2, 3) - mat(2, 2) * mat(0, 3);
        const T a_0123 = mat(0, 2) * mat(1, 3) - mat(1, 2) * mat(0, 3);
        const T a_2313 = mat(2, 1) * mat(3, 3) - mat(3, 1) * mat(2, 3);
        const T a_1313 = mat(1, 1) * mat(3, 3) - mat(3, 1) * mat(1, 3);
        const T a_1213 = mat(1, 1) * mat(2, 3) - mat(2, 1) * mat(1, 3);
        const T a_2312 = mat(2, 1) * mat(3, 2) - mat(3, 1) * mat(2, 2);
        const T a_1312 = mat(1, 1) * mat(3, 2) - mat(3, 1) * mat(1, 2);
        const T a_1212 = mat(1, 1) * mat(2, 2) - mat(2, 1) * mat(1, 2);
        const T a_0313 = mat(0, 1) * mat(3, 3) - mat(3, 1) * mat(0, 3);
        const T a_0213 = mat(0, 1) * mat(2, 3) - mat(2, 1) * mat(0, 3);
        const T a_0312 = mat(0, 1) * mat(3, 2) - mat(3, 1) * mat(0, 2);
        const T a_0212 = mat(0, 1) * mat(2, 2) - mat(2, 1) * mat(0, 2);
        const T a_0113 = mat(0, 1) * mat(1, 3) - mat(1, 1) * mat(0, 3);
        const T a_0112 = mat(0, 1) * mat(1, 2) - mat(1, 1) * mat(0, 2);

        Matrix4x4<T> result;
        result(0, 0) = +(mat(1, 1) * a_2323 - mat(2, 1) * a_1323 + mat(3, 1) * a_1223) * determinant;
        result(1, 0) = -(mat(1, 0) * a_2323 - mat(2, 0) * a_1323 + mat(3, 0) * a_1223) * determinant;
        result(2, 0) = +(mat(1, 0) * a_2313 - mat(2, 0) * a_1313 + mat(3, 0) * a_1213) * determinant;
        result(3, 0) = -(mat(1, 0) * a_2312 - mat(2, 0) * a_1312 + mat(3, 0) * a_1212) * determinant;
        result(0, 1) = -(mat(0, 1) * a_2323 - mat(2, 1) * a_0323 + mat(3, 1) * a_0223) * determinant;
        result(1, 1) = +(mat(0, 0) * a_2323 - mat(2, 0) * a_0323 + mat(3, 0) * a_0223) * determinant;
        result(2, 1) = -(mat(0, 0) * a_2313 - mat(2, 0) * a_0313 + mat(3, 0) * a_0213) * determinant;
        result(3, 1) = +(mat(0, 0) * a_2312 - mat(2, 0) * a_0312 + mat(3, 0) * a_0212) * determinant;
        result(0, 2) = +(mat(0, 1) * a_1323 - mat(1, 1) * a_0323 + mat(3, 1) * a_0123) * determinant;
        result(1, 2) = -(mat(0, 0) * a_1323 - mat(1, 0) * a_0323 + mat(3, 0) * a_0123) * determinant;
        result(2, 2) = +(mat(0, 0) * a_1313 - mat(1, 0) * a_0313 + mat(3, 0) * a_0113) * determinant;
        result(3, 2) = -(mat(0, 0) * a_1312 - mat(1, 0) * a_0312 + mat(3, 0) * a_0112) * determinant;
        result(0, 3) = -(mat(0, 1) * a_1223 - mat(1, 1) * a_0223 + mat(2, 1) * a_0123) * determinant;
        result(1, 3) = +(mat(0, 0) * a_1223 - mat(1, 0) * a_0223 + mat(2, 0) * a_0123) * determinant;
        result(2, 3) = -(mat(0, 0) * a_1213 - mat(1, 0) * a_0213 + mat(2, 0) * a_0113) * determinant;
        result(3, 3) = +(mat(0, 0) * a_1212 - mat(1, 0) * a_0212 + mat(2, 0) * a_0112) * determinant;
        return result;
    }

    template<typename T>
    constexpr Matrix4x4<T> transpose(const Matrix4x4<T>& mat)
    {
        Matrix4x4<T> result;
        for (int y = 0; y < 4; y++) {
            for (int x = 0; x < 4; x++) {
                result(x, y) = mat(y, x);
            }
        }
        return result;
    }

    // Coords
    template<typename T>
    constexpr Vector2<T> calc_ndc(const Vector2<T>& position, const Vector2<T>& frame_size)
    {
        const Vector2<T> result = {
            position.x / frame_size.x,
            (frame_size.y - position.y) / frame_size.y,
        };
        return result * T(2) - Vector2<T>(T(1));
    }

    template<typename T>
    constexpr Vector2<T> calc_ndc_ar(const Vector2<T>& position, const Vector2<T>& size)
    {
        Vector2<T> result = calc_ndc(position, size);
        result.x *= size.x / size.y;
        return result;
    }

    // Lines
    template<typename T>
    constexpr float line_x(const Vector2<T>& a, const Vector2<T>& b, T y)
    {
        return ((y - a.y) * (b.x - a.x)) / (b.y - a.y) + a.x;
    }

    template<typename T>
    constexpr float line_y(const Vector2<T>& a, const Vector2<T>& b, T x)
    {
        return ((b.y - a.y) * (x - a.x)) / (b.x - a.x) + a.y;
    }

    // Rotation
    template<typename T>
    constexpr Quaternion<T> to_quaternion(const Vector3<T>& euler)
    {
        const T cr = cos_deg(euler.x * T(0.5));
        const T sr = sin_deg(euler.x * T(0.5));
        const T cp = cos_deg(euler.y * T(0.5));
        const T sp = sin_deg(euler.y * T(0.5));
        const T cy = cos_deg(euler.z * T(0.5));
        const T sy = sin_deg(euler.z * T(0.5));
        return {
            cr * cp * cy + sr * sp * sy,
            sr * cp * cy - cr * sp * sy,
            cr * sp * cy + sr * cp * sy,
            cr * cp * sy - sr * sp * cy,
        };
    }

    template<typename T>
    constexpr Vector3<T> to_euler(const Quaternion<T>& quat)
    {
        const T       sin_p = T(+2) * (quat.w * quat.y - quat.z * quat.x) + T(0);
        const T sin_r_cos_p = T(+2) * (quat.w * quat.x + quat.y * quat.z) + T(0);
        const T cos_r_cos_p = T(-2) * (quat.x * quat.x + quat.y * quat.y) + T(1);
        const T sin_y_cos_p = T(+2) * (quat.w * quat.z + quat.x * quat.y) + T(0);
        const T cos_y_cos_p = T(-2) * (quat.y * quat.y + quat.z * quat.z) + T(1);

        const Vector3<T> result = {
            std::atan2(sin_r_cos_p, cos_r_cos_p),
            std::abs(sin_p) >= T(1) ? (T) std::copysign(PI * T(0.5), sin_p) : (T) std::asin(sin_p),
            std::atan2(sin_y_cos_p, cos_y_cos_p),
        };
        return result * (T) TO_DEGREES;
    }

    template<typename T>
    constexpr Quaternion<T> to_quaternion(const Vector3<T>& original, const Vector3<T>& target)
    {
        const Vector3<T> axis = normalize(cross(original, target));
        const T angl = angle(original, target);
        return Quaternion{ axis, angl };
    }

    template<typename T>
    constexpr Vector3<T> to_euler(const Vector3<T>& original, const Vector3<T>& target)
    {
        const Quaternion<T> quat = to_quaternion(original, target);
        return to_euler(quat);
    }
}
