#include "app.h"


App::App()
{
	window.on_resize.emplace_back([&](const kl::Int2 size)
	{
		if (size.x <= 0 || size.y <= 0)
			return;
		gpu.resize_internal(size);
		gpu.set_viewport_size(size);
	});
	window.set_dark_mode(true);
	window.maximize();

	ImGui::CreateContext();
	ImGui::GetIO().ConfigFlags |= ImGuiConfigFlags_DockingEnable;
	ImGui_ImplWin32_Init(static_cast<HWND>(window));
	ImGui_ImplDX11_Init(gpu.device().Get(), gpu.context().Get());
	dpi_scaling = ImGui_ImplWin32_GetDpiScaleForHwnd(static_cast<HWND>(window));

	load_custom_fonts();
	load_custom_theme();

	simulation_image.load_from_file("resource/simulation.png");
	simulation_view = gpu.create_shader_view(gpu.create_texture(simulation_image), nullptr);

	kl::console::set_enabled(false);
}

bool App::update()
{
	gpu.clear_internal(kl::colors::GRAY);

	ImGui_ImplWin32_NewFrame();
	ImGui_ImplDX11_NewFrame();
	ImGui::NewFrame();
	ImGuizmo::BeginFrame();
	ImGui::DockSpaceOverViewport(0, nullptr, ImGuiDockNodeFlags_PassthruCentralNode);

	render_ui();

	ImGui::Render();
	ImGui_ImplDX11_RenderDrawData(ImGui::GetDrawData());

	gpu.swap_buffers(true);
	return window.process(false);
}

void App::render_ui()
{
	ImGui::PushStyleColor(ImGuiCol_WindowBg, (ImU32) ImColor(255, 255, 255));
	if (ImGui::Begin("Simulation")) {
		static const ImVec2 button_size{ 90.0f, 30.0f };
		static const ImVec2 add_pressure_location{ 720.0f, 80.0f };
		static const ImVec2 release_pressure_location{ 40.0f, 1220.0f };
		static const ImVec2 left_box_tl{ 798.0f, 946.0f };
		static const ImVec2 right_box_tr{ 1057.0f, 946.0f };
		static const ImVec2 middle_box{ 915.0f, 1339.0f };
		static const float one_perc_height = 16.5f;

		const ImVec2 region_size = ImGui::GetContentRegionAvail();
		ImVec2 image_scale{ (float) simulation_image.width(), (float) simulation_image.height() };
		image_scale *= std::min(region_size.x / image_scale.x, region_size.y / image_scale.y);
		const ImVec2 image_cursor_pos = ImGui::GetCursorPos() + (region_size - image_scale) * 0.5f;

		const auto compute_window_point = [&](const ImVec2& image_point)
		{
			const ImVec2 original_size{ (float) simulation_image.width(), (float) simulation_image.height() };
			return (image_point / original_size) * image_scale + image_cursor_pos;
		};

		const ImVec2 win_pos = ImGui::GetWindowPos();
		auto draw_list = ImGui::GetWindowDrawList();
		draw_list->AddRectFilled(
			win_pos + compute_window_point(left_box_tl + ImVec2{ 0.0, one_perc_height * current_pressure }),
			win_pos + compute_window_point(middle_box),
			ImColor(65, 135, 245));
		draw_list->AddRectFilled(
			win_pos + compute_window_point(ImVec2{ middle_box.x, right_box_tr.y - one_perc_height * current_pressure }),
			win_pos + compute_window_point(ImVec2{ right_box_tr.x, middle_box.y }),
			ImColor(65, 135, 245));

		ImGui::SetCursorPos(image_cursor_pos);
		ImGui::Image(simulation_view.Get(), image_scale, { 0.0f, 1.0f }, { 1.0f, 0.0f });
		
		ImGui::SetCursorPos(compute_window_point(add_pressure_location) - button_size * 0.5f);
		if (ImGui::Button("ADD", button_size)) {
			current_pressure += 1.0f;
			current_pressure = std::min(current_pressure, 16.0f);
		}

		ImGui::SetCursorPos(compute_window_point(release_pressure_location) - button_size * 0.5f);
		if (ImGui::Button("RELEASE", button_size)) {
			const float h1 = current_pressure;
			const float h2 = h1 - (h1 / (1.005f / 0.718f));
			current_pressure = -h2;
			current_pressure = std::max(current_pressure, -16.0f);
		}
	}
	ImGui::End();
	ImGui::PopStyleColor(1);

	if (ImGui::Begin("Data")) {
		float y_average = 0.0f;
		int sample_count = 0;

		ImGui::PushStyleVar(ImGuiStyleVar_CellPadding, ImVec2(0.0f, 5.0f));
		if (ImGui::BeginTable("##DataTable", 3)) {
			ImGui::TableNextColumn();
			ImGui::Text("h1");
			ImGui::TableNextColumn();
			ImGui::Text("h2");
			ImGui::TableNextColumn();
			ImGui::Text("cp/cv");

			for (int i = 0; i < (int) pressure_data.size(); i++) {
				float y_value = pressure_data[i].first / (pressure_data[i].first - pressure_data[i].second);
				if (pressure_data[i].first || pressure_data[i].second) {
					y_average += y_value;
					sample_count += 1;
				}

				ImGui::TableNextColumn();
				ImGui::DragFloat(kl::format("##h1_", i).c_str(), &pressure_data[i].first, 0.1f);
				ImGui::TableNextColumn();
				ImGui::DragFloat(kl::format("##h2_", i).c_str(), &pressure_data[i].second, 0.1f);
				ImGui::TableNextColumn();
				ImGui::DragFloat(kl::format("##y_", i).c_str(), &y_value, 0.0f);
			}
			ImGui::EndTable();
		}
		ImGui::PopStyleVar(1);

		y_average /= sample_count;

		float y_error = 0.0f;
		for (int i = 0; i < (int) pressure_data.size(); i++) {
			if (pressure_data[i].first || pressure_data[i].second) {
				const float y_value = pressure_data[i].first / (pressure_data[i].first - pressure_data[i].second);
				y_error += std::pow(y_value - y_average, 2.0f);
			}
		}
		y_error = std::sqrt(y_error / (sample_count - 1));

		ImGui::Text(kl::format("srednja vrijednost: ", y_average).c_str());
		ImGui::Text(kl::format("pogreska mjerenja: ", y_error).c_str());
		ImGui::Text(kl::format("ukupno: ", y_average, " +- ", y_error).c_str());
	}
	ImGui::End();
}

void App::load_custom_fonts()
{
	ImFontAtlas* atlas = ImGui::GetIO().Fonts;
	auto font = atlas->AddFontFromFileTTF("resource/roboto.ttf", 16 * dpi_scaling);
	kl::assert(font, "Failed to load the font");
	atlas->Build();
}

void App::load_custom_theme()
{
	const kl::Float4 special_color = kl::colors::WHITE;

	ImGuiStyle& style = ImGui::GetStyle();

	style.Colors[ImGuiCol_Text] = ImVec4(1.00f, 1.00f, 1.00f, 1.00f);
	style.Colors[ImGuiCol_TextDisabled] = ImVec4(0.50f, 0.50f, 0.50f, 1.00f);

	style.Colors[ImGuiCol_WindowBg] = ImVec4(0.10f, 0.10f, 0.10f, 1.00f);
	style.Colors[ImGuiCol_ChildBg] = ImVec4(0.14f, 0.14f, 0.14f, 1.00f);
	style.Colors[ImGuiCol_PopupBg] = ImVec4(0.14f, 0.14f, 0.14f, 1.00f);

	style.Colors[ImGuiCol_Border] = ImVec4(0.45f, 0.45f, 0.45f, 0.50f);
	style.Colors[ImGuiCol_BorderShadow] = ImVec4(0.00f, 0.00f, 0.00f, 0.00f);

	style.Colors[ImGuiCol_FrameBg] = ImVec4(0.15f, 0.15f, 0.15f, 1.00f);
	style.Colors[ImGuiCol_FrameBgHovered] = ImVec4(0.30f, 0.30f, 0.30f, 1.00f);
	style.Colors[ImGuiCol_FrameBgActive] = ImVec4(0.60f, 0.60f, 0.60f, 0.40f);

	style.Colors[ImGuiCol_TitleBg] = ImVec4(0.08f, 0.08f, 0.08f, 1.00f);
	style.Colors[ImGuiCol_TitleBgActive] = ImVec4(0.08f, 0.08f, 0.08f, 1.00f);
	style.Colors[ImGuiCol_TitleBgCollapsed] = ImVec4(0.00f, 0.00f, 0.00f, 0.51f);

	style.Colors[ImGuiCol_MenuBarBg] = ImVec4(0.14f, 0.14f, 0.14f, 1.00f);

	style.Colors[ImGuiCol_ScrollbarBg] = ImVec4(0.02f, 0.02f, 0.02f, 0.53f);
	style.Colors[ImGuiCol_ScrollbarGrab] = ImVec4(0.31f, 0.31f, 0.31f, 1.00f);
	style.Colors[ImGuiCol_ScrollbarGrabHovered] = ImVec4(0.41f, 0.41f, 0.41f, 1.00f);
	style.Colors[ImGuiCol_ScrollbarGrabActive] = ImVec4(0.51f, 0.51f, 0.51f, 1.00f);

	style.Colors[ImGuiCol_CheckMark] = (const ImVec4&) special_color;

	style.Colors[ImGuiCol_SliderGrab] = (const ImVec4&) special_color;
	style.Colors[ImGuiCol_SliderGrabActive] = (const ImVec4&) special_color;

	style.Colors[ImGuiCol_Button] = style.Colors[ImGuiCol_FrameBg];
	style.Colors[ImGuiCol_ButtonHovered] = style.Colors[ImGuiCol_FrameBgHovered];
	style.Colors[ImGuiCol_ButtonActive] = style.Colors[ImGuiCol_FrameBgActive];

	style.Colors[ImGuiCol_Header] = ImVec4(0.22f, 0.22f, 0.22f, 1.00f);
	style.Colors[ImGuiCol_HeaderHovered] = ImVec4(0.25f, 0.25f, 0.25f, 1.00f);
	style.Colors[ImGuiCol_HeaderActive] = ImVec4(0.67f, 0.67f, 0.67f, 0.39f);

	style.Colors[ImGuiCol_Separator] = style.Colors[ImGuiCol_Border];
	style.Colors[ImGuiCol_SeparatorHovered] = ImVec4(0.44f, 0.44f, 0.44f, 1.00f);
	style.Colors[ImGuiCol_SeparatorActive] = (const ImVec4&) special_color;

	style.Colors[ImGuiCol_ResizeGrip] = ImVec4(0.00f, 0.00f, 0.00f, 0.00f);
	style.Colors[ImGuiCol_ResizeGripHovered] = ImVec4(0.30f, 0.30f, 0.30f, 0.67f);
	style.Colors[ImGuiCol_ResizeGripActive] = (const ImVec4&) special_color;

	style.Colors[ImGuiCol_Tab] = ImVec4(0.08f, 0.08f, 0.08f, 0.83f);
	style.Colors[ImGuiCol_TabHovered] = ImVec4(0.35f, 0.35f, 0.35f, 0.83f);
	style.Colors[ImGuiCol_TabActive] = ImVec4(0.23f, 0.23f, 0.23f, 1.00f);
	style.Colors[ImGuiCol_TabUnfocused] = ImVec4(0.08f, 0.08f, 0.08f, 1.00f);
	style.Colors[ImGuiCol_TabUnfocusedActive] = ImVec4(0.14f, 0.14f, 0.14f, 1.00f);

	style.Colors[ImGuiCol_DockingPreview] = (const ImVec4&) special_color;
	style.Colors[ImGuiCol_DockingEmptyBg] = ImVec4(0.20f, 0.20f, 0.20f, 1.00f);

	style.Colors[ImGuiCol_PlotLines] = ImVec4(0.61f, 0.61f, 0.61f, 1.00f);
	style.Colors[ImGuiCol_PlotLinesHovered] = ImVec4(1.00f, 0.43f, 0.35f, 1.00f);
	style.Colors[ImGuiCol_PlotHistogram] = ImVec4(0.90f, 0.70f, 0.00f, 1.00f);
	style.Colors[ImGuiCol_PlotHistogramHovered] = ImVec4(1.00f, 0.60f, 0.00f, 1.00f);

	style.Colors[ImGuiCol_TextSelectedBg] = ImVec4(0.42f, 0.69f, 0.69f, 0.32f);

	style.Colors[ImGuiCol_DragDropTarget] = (const ImVec4&) special_color;

	style.Colors[ImGuiCol_NavHighlight] = (const ImVec4&) special_color;
	style.Colors[ImGuiCol_NavWindowingHighlight] = ImVec4(1.00f, 1.00f, 1.00f, 0.70f);
	style.Colors[ImGuiCol_NavWindowingDimBg] = ImVec4(0.80f, 0.80f, 0.80f, 0.20f);

	style.Colors[ImGuiCol_ModalWindowDimBg] = ImVec4(0.80f, 0.80f, 0.80f, 0.35f);

	style.WindowPadding = ImVec2(8.00f, 8.00f);
	style.FramePadding = ImVec2(5.00f, 2.00f);
	style.CellPadding = ImVec2(6.00f, 6.00f);
	style.ItemSpacing = ImVec2(6.00f, 6.00f);
	style.ItemInnerSpacing = ImVec2(6.00f, 6.00f);
	style.TouchExtraPadding = ImVec2(0.00f, 0.00f);
	style.IndentSpacing = 25;
	style.ScrollbarSize = 15;
	style.GrabMinSize = 10;
	style.WindowBorderSize = 1;
	style.ChildBorderSize = 1;
	style.PopupBorderSize = 1;
	style.FrameBorderSize = 1;
	style.TabBorderSize = 1;
	style.WindowRounding = 7;
	style.ChildRounding = 4;
	style.FrameRounding = 3;
	style.PopupRounding = 4;
	style.ScrollbarRounding = 9;
	style.GrabRounding = 3;
	style.LogSliderDeadzone = 4;
	style.TabRounding = 4;
}
